@@ -1,4 +1,4 @@
-/* Copyright (c) 2013 Raymond Wagenmaker
+/* Copyright (c) 2013-2014 Raymond Wagenmaker
  *
  * See LICENSE for the MIT license.
  */
@@ -14,6 +14,7 @@ G_BEGIN_DECLS
 
 #define XCB_SCREENSAVER_PROPERTY_NAME "_MIT_SCREEN_SAVER_ID"
 
+#if XCB_SCREENSAVER_NOTIFY_EVENT_T_BROKEN
 #define xcb_screensaver_notify_event_t xcb_screensaver_notify_event_t_fixed
 typedef struct xcb_screensaver_notify_event_t_fixed {
     uint8_t         response_type; /**<  */
@@ -26,11 +27,10 @@ typedef struct xcb_screensaver_notify_event_t_fixed {
     uint8_t         forced; /**<  */
     uint8_t         pad1[14]; /**<  */
 } xcb_screensaver_notify_event_t_fixed;
+#endif
 
 GQuark xcb_error_quark(void) G_GNUC_CONST;
 
-xcb_screen_t *xcb_get_screen(xcb_connection_t *connection, int screen_number);
-
 typedef gboolean (*XcbEventFunc)(xcb_connection_t *connection, xcb_generic_event_t *event, gpointer user_data);
 
 GSource *xcb_event_source_new(xcb_connection_t *connection);
