@@ -1,7 +1,8 @@
-/* Copyright (c) 2013 Raymond Wagenmaker
+/* Copyright (c) 2013-2014 Raymond Wagenmaker
  *
  * See LICENSE for the MIT license.
  */
+#include <assert.h>
 #include <locale.h>
 #include <fcntl.h>
 #include <stdlib.h>
@@ -12,6 +13,7 @@
 #include <gio/gio.h>
 #include <gio/gunixfdlist.h>
 #include <xcb/xcb.h>
+#include <xcb/xcb_aux.h>
 #include <xcb/xcb_event.h>
 #include <xcb/screensaver.h>
 
@@ -35,8 +37,10 @@ static gboolean register_screensaver(xcb_connection_t *connection, xcb_screen_t
 static void unregister_screensaver(xcb_connection_t *connection, xcb_screen_t *screen, xcb_atom_t atom);
 static gboolean screensaver_event_cb(xcb_connection_t *connection, xcb_generic_event_t *event, const int *xcb_screensaver_notify);
 
+static int get_cycle_time(xcb_connection_t *connection);
 static void keep_sleep_lock_fd_open(gpointer user_data);
 static void start_child(Child *child);
+static void kill_child_signal(Child *child, int signal);
 static void kill_child(Child *child);
 static void child_watch_cb(GPid pid, gint status, Child *child);
 
@@ -48,6 +52,7 @@ static void logind_manager_call_get_session_cb(GObject *source_object, GAsyncRes
 static void logind_session_proxy_new_cb(GObject *source_object, GAsyncResult *res, gpointer user_data);
 static void logind_session_on_signal_lock(GDBusProxy *proxy, gchar *sender_name, gchar *signal_name, GVariant *parameters, gpointer user_data);
 static void logind_session_set_idle_hint(gboolean idle);
+static void logind_session_set_locked_hint(gboolean locked);
 
 static gboolean parse_options(int argc, char *argv[], GError **error);
 static gboolean parse_notifier_cmd(const gchar *option_name, const gchar *value, gpointer data, GError **error);
@@ -60,25 +65,26 @@ static Child locker = {"locker", NULL, 0, FALSE, &notifier};
 static gboolean opt_quiet = FALSE;
 static gboolean opt_verbose = FALSE;
 static gboolean opt_ignore_sleep = FALSE;
+static gboolean opt_ignore_xss = FALSE;
 static gboolean opt_print_version = FALSE;
+static gchar *opt_session = NULL;
 
 static GOptionEntry opt_entries[] = {
     {G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &locker.cmd, NULL, "LOCK_CMD [ARG...]"},
     {"notifier", 'n', G_OPTION_FLAG_FILENAME, G_OPTION_ARG_CALLBACK, parse_notifier_cmd, "Send notification using CMD", "CMD"},
     {"transfer-sleep-lock", 'l', 0, G_OPTION_ARG_NONE, &locker.transfer_sleep_lock_fd, "Pass sleep delay lock file descriptor to locker", NULL},
     {"ignore-sleep", 0, 0, G_OPTION_ARG_NONE, &opt_ignore_sleep, "Do not lock on suspend/hibernate", NULL},
+    {"ignore-xss", 0, 0, G_OPTION_ARG_NONE, &opt_ignore_xss, "Do not lock on X screen saver events", NULL},
     {"quiet", 'q', 0, G_OPTION_ARG_NONE, &opt_quiet, "Output only fatal errors", NULL},
     {"verbose", 'v', 0, G_OPTION_ARG_NONE, &opt_verbose, "Output more messages", NULL},
     {"version", 0, 0, G_OPTION_ARG_NONE, &opt_print_version, "Print version number and exit", NULL},
+    {"session", 's', 0, G_OPTION_ARG_STRING, &opt_session, "Use ID instead of the current session", "ID"},
     {NULL}
 };
 
 static GDBusProxy *logind_manager = NULL;
-
 static GDBusProxy *logind_session = NULL;
-
 static gint sleep_lock_fd = -1;
-
 static gboolean preparing_for_sleep = FALSE;
 
 static gboolean
@@ -158,13 +164,19 @@ static gboolean
 screensaver_event_cb(xcb_connection_t *connection, xcb_generic_event_t *event,
                      const int *const xcb_screensaver_notify)
 {
-    const uint8_t type = XCB_EVENT_RESPONSE_TYPE(event);
+    uint8_t event_type;
 
-    if (type == 0) {
+    if (!event) {
+        g_info("X connection lost; exiting.");
+        exit(EXIT_FAILURE);
+    }
+
+    event_type = XCB_EVENT_RESPONSE_TYPE(event);
+    if (event_type == 0) {
         xcb_generic_error_t *error = (xcb_generic_error_t *)event;
 
         g_warning("X error: %s", xcb_event_get_error_label(error->error_code));
-    } else if (type == *xcb_screensaver_notify) {
+    } else if (event_type == *xcb_screensaver_notify) {
         xcb_screensaver_notify_event_t *xss_event =
             (xcb_screensaver_notify_event_t *)event;
 
@@ -178,8 +190,12 @@ screensaver_event_cb(xcb_connection_t *connection, xcb_generic_event_t *event,
                  * work that way; I'm leaving this in anyway.
                  */
                 xcb_force_screen_saver(connection, XCB_SCREEN_SAVER_ACTIVE);
-            else if (!notifier.cmd || xss_event->forced) {
-                start_child(&locker);
+            /* skip notifier if it's not defined, screen saver is forced or if cycle time is 0 */
+            else if (!notifier.cmd || xss_event->forced || get_cycle_time(connection) <= 0) {
+                if (!opt_ignore_xss) {
+                    start_child(&locker);
+                    logind_session_set_locked_hint(TRUE);
+                }
                 logind_session_set_idle_hint(TRUE);
             } else if (!locker.pid)
                 start_child(&notifier);
@@ -187,13 +203,16 @@ screensaver_event_cb(xcb_connection_t *connection, xcb_generic_event_t *event,
                 logind_session_set_idle_hint(TRUE);
             break;
         case XCB_SCREENSAVER_STATE_OFF:
-            kill_child(&notifier);
+            kill_child_signal(&notifier, SIGHUP);
             logind_session_set_idle_hint(FALSE);
             break;
         case XCB_SCREENSAVER_STATE_CYCLE:
             if (!locker.pid) {
                 logind_session_set_idle_hint(TRUE);
-                start_child(&locker);
+                if (!opt_ignore_xss) {
+                    start_child(&locker);
+                    logind_session_set_locked_hint(TRUE);
+                }
             }
             break;
         }
@@ -201,6 +220,26 @@ screensaver_event_cb(xcb_connection_t *connection, xcb_generic_event_t *event,
     return TRUE;
 }
 
+static int
+get_cycle_time(xcb_connection_t *connection)
+{
+    xcb_get_screen_saver_cookie_t cookie;
+    xcb_get_screen_saver_reply_t *reply;
+    xcb_generic_error_t *error;
+    int cycle_time;
+
+    cookie = xcb_get_screen_saver(connection);
+    if ((reply = xcb_get_screen_saver_reply(connection, cookie, &error))) {
+        cycle_time = (int) reply->interval;
+        free(reply);
+        return cycle_time;
+    } else {
+        g_warning("X error: %s", xcb_event_get_error_label(error->error_code));
+        free(error);
+        return -1;
+    }
+}
+
 static void
 keep_sleep_lock_fd_open(gpointer user_data)
 {
@@ -241,26 +280,41 @@ out:
     g_strfreev(env);
 }
 
+static void
+kill_child_signal(Child *child, int signal)
+{
+    if (child->pid && kill(child->pid, signal))
+        g_warning("Error sending %s to %s: %s", strsignal(signal), child->name, g_strerror(errno));
+}
+
 static void
 kill_child(Child *child)
 {
-    if (child->pid && kill(child->pid, SIGTERM))
-        g_warning("Error sending SIGTERM to %s: %s", child->name, g_strerror(errno));
+    kill_child_signal(child, SIGTERM);
 }
 
 static void
 child_watch_cb(GPid pid, gint status, Child *child)
 {
 #if GLIB_CHECK_VERSION(2, 34, 0)
+    gboolean success;
     GError *error = NULL;
 
-    if (!g_spawn_check_exit_status(status, &error)) {
+#if GLIB_CHECK_VERSION(2, 70, 0)
+    success = g_spawn_check_wait_status(status, &error);
+#else
+    // Use the old deprecated name
+    success = g_spawn_check_exit_status(status, &error);
+#endif
+
+    if (!success) {
         g_message("%s exited abnormally: %s", child->name, error->message);
         g_error_free(error);
     }
 #endif
     child->pid = 0;
     g_spawn_close_pid(pid);
+    if (child == &locker) logind_session_set_locked_hint(FALSE);
 }
 
 static void
@@ -281,8 +335,18 @@ logind_manager_proxy_new_cb(GObject *source_object, GAsyncResult *res,
                          G_CALLBACK(logind_manager_on_signal_prepare_for_sleep), NULL);
         logind_manager_take_sleep_delay_lock();
     }
-    g_dbus_proxy_call(logind_manager, "GetSessionByPID",
-                      g_variant_new("(u)", getpid()), G_DBUS_CALL_FLAGS_NONE,
+
+    GVariant *data;
+    gchar *name;
+    if (!user_data) {
+        data = g_variant_new("(u)", getpid());
+        name = "GetSessionByPID";
+    } else {
+        data = g_variant_new("(s)", user_data);
+        name = "GetSession";
+    }
+
+    g_dbus_proxy_call(logind_manager, name, data, G_DBUS_CALL_FLAGS_NONE,
                       -1, NULL, logind_manager_call_get_session_cb, NULL);
 }
 
@@ -344,6 +408,7 @@ logind_manager_on_signal_prepare_for_sleep(GDBusProxy *proxy,
         preparing_for_sleep = TRUE;
 
         start_child(&locker);
+        logind_session_set_locked_hint(TRUE);
 
         if (sleep_lock_fd >= 0) {
             close(sleep_lock_fd);
@@ -364,7 +429,7 @@ logind_manager_call_get_session_cb(GObject *source_object, GAsyncResult *res,
 
     result = g_dbus_proxy_call_finish(logind_manager, res, &error);
     if (!result) {
-        g_warning("Error getting current session: %s", error->message);
+        g_warning("Error getting session: %s", error->message);
         g_error_free(error);
         return;
     }
@@ -402,18 +467,29 @@ logind_session_on_signal_lock(GDBusProxy *proxy,
                               GVariant   *parameters,
                               gpointer    user_data)
 {
-    if (!g_strcmp0(signal_name, "Lock"))
+    if (!g_strcmp0(signal_name, "Lock")) {
         start_child(&locker);
+        logind_session_set_locked_hint(TRUE);
+    }
     else if (!g_strcmp0(signal_name, "Unlock"))
         kill_child(&locker);
 }
 
+static void
+logind_session_set_locked_hint(gboolean locked)
+{
+     assert(logind_session);
+     g_dbus_proxy_call(logind_session, "SetLockedHint",
+                       g_variant_new("(b)", locked),
+                       G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
+}
+
 static void
 logind_session_set_idle_hint(gboolean idle)
 {
-    if (logind_session)
-        g_dbus_proxy_call(logind_session, "SetIdleHint", g_variant_new("(b)", idle),
-                          G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
+    assert(logind_session);
+    g_dbus_proxy_call(logind_session, "SetIdleHint", g_variant_new("(b)", idle),
+                      G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
 }
 
 static gboolean
@@ -456,7 +532,6 @@ reset_screensaver(xcb_connection_t *connection)
 {
     if (!locker.pid)
         xcb_force_screen_saver(connection, XCB_SCREEN_SAVER_RESET);
-
     return TRUE;
 }
 
@@ -474,7 +549,7 @@ log_handler(const gchar *log_domain, GLogLevelFlags log_level,
             const gchar *message, gpointer user_data)
 {
     if (opt_verbose || log_level & G_LOG_FLAG_FATAL
-    || !(opt_quiet || log_level & G_LOG_LEVEL_MESSAGE))
+        || !(opt_quiet || log_level & G_LOG_LEVEL_MESSAGE))
         g_log_default_handler(log_domain, log_level, message, user_data);
 }
 
@@ -498,6 +573,7 @@ main(int argc, char *argv[])
         goto init_error;
     }
     g_log_set_default_handler(log_handler, NULL);
+    g_log_set_fatal_mask(NULL, G_LOG_LEVEL_CRITICAL);
 
     connection = xcb_connect(NULL, &default_screen_number);
     if (xcb_connection_has_error(connection)) {
@@ -512,14 +588,14 @@ main(int argc, char *argv[])
     g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
                              G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES, NULL,
                              LOGIND_SERVICE, LOGIND_PATH, LOGIND_MANAGER_INTERFACE,
-                             NULL, logind_manager_proxy_new_cb, NULL);
+                             NULL, logind_manager_proxy_new_cb, opt_session);
 
     loop = g_main_loop_new(NULL, FALSE);
     g_unix_signal_add(SIGTERM, (GSourceFunc)exit_service, loop);
-    g_unix_signal_add(SIGINT, (GSourceFunc)exit_service, loop);
-    g_unix_signal_add(SIGHUP, (GSourceFunc)reset_screensaver, connection);
+    g_unix_signal_add(SIGINT,  (GSourceFunc)exit_service, loop);
+    g_unix_signal_add(SIGHUP,  (GSourceFunc)exit_service, loop);
 
-    default_screen = xcb_get_screen(connection, default_screen_number);
+    default_screen = xcb_aux_get_screen(connection, default_screen_number);
     if (!register_screensaver(connection, default_screen, &atom, &error))
         goto init_error;
 
